@@ -1,12 +1,14 @@
 from __future__ import annotations
-from typing import List, Tuple
-from staticsite import Page, Feature, File
+from typing import List, Tuple, Optional
+from staticsite import Page, Feature
+from staticsite.page import PageNotFoundError
 from staticsite.utils import front_matter
 from staticsite.archetypes import Archetype
-from staticsite.contents import ContentDir, Dir
-from staticsite.utils import lazy
+from staticsite.contents import ContentDir
 from staticsite.utils.typing import Meta
+from urllib.parse import urlparse, urlunparse
 import jinja2
+import re
 import os
 import io
 import markdown
@@ -19,33 +21,38 @@ class LinkResolver(markdown.treeprocessors.Treeprocessor):
     def __init__(self, *args, **kw):
         super().__init__(*args, **kw)
         self.page = None
+        self.absolute = False
         self.substituted = {}
 
-    def set_page(self, page):
+    def set_page(self, page, absolute=False):
         self.page = page
+        self.absolute = absolute
         self.substituted = {}
 
     def run(self, root):
         for a in root.iter("a"):
-            new_url = self.resolve_url(a.attrib.get("href", None))
+            page, new_url = self.resolve_url(a.attrib.get("href", None))
             if new_url is not None:
                 a.attrib["href"] = new_url
 
-        for a in root.iter("img"):
-            new_url = self.resolve_url(a.attrib.get("src", None))
-            if new_url is not None:
-                a.attrib["src"] = new_url
+        for img in root.iter("img"):
+            orig_url = img.attrib.get("src", None)
+            if orig_url is None:
+                continue
 
-    def resolve_url(self, url):
-        """
-        Resolve internal URLs.
+            page, parsed = self.resolve_page(orig_url)
+            if page is None:
+                continue
 
-        Returns None if the URL does not need changing, else returns the new URL.
-        """
-        new_url = self.substituted.get(url)
-        if new_url is not None:
-            return new_url
+            try:
+                attrs = self.page.get_img_attributes(page, absolute=self.absolute)
+            except PageNotFoundError as e:
+                log.warn("%s: %s", self.page, e)
+                continue
 
+            img.attrib.update(attrs)
+
+    def resolve_page(self, url) -> Tuple[Page, Tuple]:
         from markdown.util import AMP_SUBSTITUTE
         if url.startswith(AMP_SUBSTITUTE):
             # Possibly an overencoded mailto: link.
@@ -53,14 +60,56 @@ class LinkResolver(markdown.treeprocessors.Treeprocessor):
             #
             # Markdown then further escapes & with utils.AMP_SUBSTITUTE, so
             # we look for it here.
-            return None
+            return None, None
 
-        new_url = self.page.resolve_url(url)
-        if new_url is url:
-            return None
+        parsed = urlparse(url)
 
-        self.substituted[url] = new_url
-        return new_url
+        # If it's an absolute url, leave it unchanged
+        if parsed.scheme or parsed.netloc:
+            return None, parsed
+
+        # If it's an anchor inside the page, leave it unchanged
+        if not parsed.path:
+            return None, parsed
+
+        # Try with cache
+        site_path = self.substituted.get(parsed.path)
+        if site_path is not None:
+            try:
+                return self.page.site.pages[site_path], parsed
+            except KeyError:
+                log.warn("%s: url %s resolved via cache to %s which does not exist in the site. Cache out of date?",
+                         self.page, url, site_path)
+
+        # Resolve as a path
+        try:
+            page = self.page.resolve_path(parsed.path)
+        except PageNotFoundError as e:
+            log.warn("%s: %s", self.page, e)
+            return None, parsed
+
+        # Cache the page site_path
+        self.substituted[url] = page.meta["site_path"]
+
+        return page, parsed
+
+    def resolve_url(self, url) -> Tuple[Page, str]:
+        """
+        Resolve internal URLs.
+
+        Returns None if the URL does not need changing, else returns the new URL.
+        """
+        page, parsed = self.resolve_page(url)
+        if page is None:
+            return page, url
+
+        new_url = self.page.url_for(page, absolute=self.absolute)
+        dest = urlparse(new_url)
+
+        return page, urlunparse(
+            (dest.scheme, dest.netloc, dest.path,
+             parsed.params, parsed.query, parsed.fragment)
+        )
 
 
 class StaticSiteExtension(markdown.extensions.Extension):
@@ -103,16 +152,23 @@ class MarkdownPages(Feature):
     def jinja2_markdown(self, context, mdtext):
         return jinja2.Markup(self.render_snippet(context.parent["page"], mdtext))
 
-    def render_page(self, page):
+    def render_page(self, page, body, render_type, absolute=False):
         """
         Render markdown in the context of the given page.
         """
-        self.link_resolver.set_page(page)
+        self.link_resolver.set_page(page, absolute)
 
-        cached = self.render_cache.get(page.src.relpath)
+        cache_key = f"{render_type}:{page.src.relpath}"
+
+        # Try fetching rendered content from cache
+        cached = self.render_cache.get(cache_key)
         if cached and cached["mtime"] != page.src.stat.st_mtime:
+            # If the source has changed, drop the cached version
             cached = None
         if cached:
+            # If the destination of links has changed, drop the cached version.
+            # This will as a side effect prime the link resolver cache,
+            # avoiding links from being looked up again during rendering
             for src, dest in cached["paths"]:
                 if self.link_resolver.resolve_url(src) != dest:
                     cached = None
@@ -121,11 +177,10 @@ class MarkdownPages(Feature):
             # log.info("%s: markdown cache hit", page.src.relpath)
             return cached["rendered"]
 
-        content = page.get_content()
         self.markdown.reset()
-        rendered = self.markdown.convert(content)
+        rendered = self.markdown.convert("\n".join(body))
 
-        self.render_cache.put(page.src.relpath, {
+        self.render_cache.put(cache_key, {
             "mtime": page.src.stat.st_mtime,
             "rendered": rendered,
             "paths": list(self.link_resolver.substituted.items()),
@@ -156,7 +211,9 @@ class MarkdownPages(Feature):
 
             meta = sitedir.meta_file(fname)
             if fname not in ("index.md", "README.md"):
-                meta["site_path"] = os.path.join(meta["site_path"], fname[:-3])
+                meta["site_path"] = os.path.join(sitedir.meta["site_path"], fname[:-3])
+            else:
+                meta["site_path"] = sitedir.meta["site_path"]
 
             try:
                 fm_meta, body = self.load_file_meta(sitedir, src, fname)
@@ -167,7 +224,7 @@ class MarkdownPages(Feature):
 
             meta.update(fm_meta)
 
-            page = MarkdownPage(self, src, meta=meta, dir=sitedir, body=body)
+            page = MarkdownPage(self.site, src, meta=meta, dir=sitedir, feature=self, body=body)
             pages.append(page)
 
         for fname in taken:
@@ -179,18 +236,17 @@ class MarkdownPages(Feature):
         # Load front matter from index.md
         # Do not try to load front matter from README.md, as one wouldn't
         # clutter a repo README with staticsite front matter
-        index = sitedir.files.get("index.md")
-        if index is None:
+        src = sitedir.files.get("index.md")
+        if src is None:
             return
 
         try:
-            with open(index.abspath, "rb") as fd:
-                fmt, meta = front_matter.read_partial(fd)
+            meta, body = self.load_file_meta(sitedir, src, "index.md")
         except Exception as e:
-            log.debug("%s: failed to parse front matter", index.relpath, exc_info=e)
-            log.warn("%s: failed to parse front matter", index.relpath)
+            log.debug("%s: failed to parse front matter", src.relpath, exc_info=e)
+            log.warn("%s: failed to parse front matter", src.relpath)
         else:
-            sitedir.add_dir_config(meta)
+            return meta
 
     def load_file_meta(self, sitedir, src, fname) -> Tuple[Meta, List[str]]:
         """
@@ -202,11 +258,9 @@ class MarkdownPages(Feature):
 
         # Parse separating front matter and markdown content
         with sitedir.open(fname, src, "rb") as fd:
-            fmt, meta = front_matter.read_partial(fd)
+            fmt, meta, body = front_matter.read_markdown_partial(fd)
 
-            body = []
-            for line in fd:
-                body.append(line.strip().decode())
+            body = list(body)
 
             # Remove leading empty lines
             while body and not body[0]:
@@ -277,7 +331,7 @@ class MarkdownArchetype(Archetype):
             # Reparse it separating front matter and markdown content
             fmatter, body = parse_markdown_with_front_matter(fd)
         try:
-            style, meta = front_matter.parse(fmatter)
+            style, meta = front_matter.read_string(fmatter)
         except Exception as e:
             log.debug("archetype %s: failed to parse front matter", self.relpath, exc_info=e)
             log.warn("archetype %s: failed to parse front matter", self.relpath)
@@ -303,30 +357,65 @@ class MarkdownArchetype(Archetype):
 class MarkdownPage(Page):
     TYPE = "markdown"
 
-    def __init__(
-            self, mdpages: MarkdownPages, src: File, meta: Meta, dir: Dir, body: List[str]):
-        super().__init__(site=mdpages.site, src=src, meta=meta, dir=dir)
+    # Match a Markdown divider line
+    re_divider = re.compile(r"^____+$")
 
-        self.meta["build_path"] = os.path.join(meta["site_path"], "index.html")
+    def __init__(self, *args, feature: MarkdownPages, body: List[str], **kw):
+        super().__init__(*args, **kw)
+
+        self.meta["build_path"] = os.path.join(self.meta["site_path"], "index.html")
 
         # Indexed by default
         self.meta.setdefault("indexed", True)
 
         # Shared markdown environment
-        self.mdpages = mdpages
+        self.mdpages = feature
 
-        # Sequence of lines found in the body
-        self.body = body
+        # Sequence of lines found in the body before the divider line, if any
+        self.body_start: List[str]
 
-    def get_content(self):
-        return "\n".join(self.body)
+        # Sequence of lines found in the body including and after the divider
+        # line, nor None if there is no divider line
+        self.body_rest: Optional[List[str]]
+
+        # Split lead and rest of the post, if a divider line is present
+        for idx, line in enumerate(body):
+            if self.re_divider.match(line):
+                self.content_has_split = True
+                self.body_start = body[:idx]
+                self.body_rest = body[idx:]
+                break
+        else:
+            self.content_has_split = False
+            self.body_start = body
+            self.body_rest = None
 
     def check(self, checker):
-        self.mdpages.render_page(self)
+        self.render()
 
-    @lazy
-    def content(self):
-        return self.mdpages.render_page(self)
+    @jinja2.contextfunction
+    def html_body(self, context, **kw) -> str:
+        if self.content_has_split:
+            body = self.body_start + ["", "<a name='sep'></a>", ""] + self.body_rest
+            return self.mdpages.render_page(self, body, render_type="hb")
+        else:
+            return self.mdpages.render_page(self, self.body_start, render_type="s")
+
+    @jinja2.contextfunction
+    def html_inline(self, context, **kw) -> str:
+        if self.content_has_split:
+            body = self.body_start + ["", f"[(continue reading)](/{self.meta['site_path']})"]
+            return self.mdpages.render_page(self, body, render_type="h")
+        else:
+            return self.mdpages.render_page(self, self.body_start, render_type="s")
+
+    @jinja2.contextfunction
+    def html_feed(self, context, **kw) -> str:
+        if self.content_has_split:
+            body = self.body_start + [""] + self.body_rest
+            return self.mdpages.render_page(self, body, render_type="f")
+        else:
+            return self.mdpages.render_page(self, self.body_start, render_type="f")
 
 
 FEATURES = {
